/* MemoServ functions.
  
   Paradox IRC Services - (c) 2001 In Mean
       E-Mail: <paradox@auspice.org> IRC: auspice@Auspice.org
   WreckedNet IRC Services - (c) 2000 Darcy Grexton
       E-Mail: <skold@wrecked.net> IRC: skold@WreckedNet
   Based on Magick 1.4
  
   Magick IRC Services are copyright (c) 1996-1998 Preston A. Elder.
       E-mail: <prez@magick.tm>   IRC: PreZ@RelicNet
   Originally based on EsperNet services (c) 1996-1998 Andy Church
   This program is free but copyrighted software; see the file COPYING for
   details.
 */

#include "services.h"
#include "ms-help.c"

MemoList *memolists[256];
MemoList *find_memolist (const char *nick);
static void alpha_insert_memolist (MemoList * ml);
static int candomemo (const char *source, const char *other);
static void do_help (const char *source);
static void do_ohelp (const char *source);
static void do_send (const char *source);
static void do_opersend (const char *source);
static void do_sopsend (const char *source);
static void do_globalmemo (const char *source);
static void do_list (const char *source);
static void do_read (const char *source);
static void do_del (const char *source);

/* Return information on memory use.  Assumes pointers are valid. */
void get_memoserv_stats (long *nrec, long *memuse)
{
    long count = 0, mem = 0;
    int i, j;
    MemoList *ml;

    for (i = 0; i < 256; i++)
	for (ml = memolists[i]; ml; ml = ml->next)
	{
	    count++;
	    mem += sizeof (*ml);
	    mem += sizeof (Memo) * ml->n_memos;
	    for (j = 0; j < ml->n_memos; j++)
		if (ml->memos[j].text)
		    mem += strlen (ml->memos[j].text) + 1;
	}
    *nrec = count;
    *memuse = mem;
}

/* Main MemoServ routine. */
void memoserv (const char *source, char *buf)
{
    char *cmd, *s;

    cmd = strtok (buf, " ");

    if (!cmd)
	return;

    else if (stricmp (cmd, "\1PING") == 0)
    {
	if (!(s = strtok (NULL, "")))
	    s = "\1";
	ctcpreply (s_MemoServ, source, "\1PING %s", s);
    }
    else if (stricmp (cmd, "\1VERSION\1") == 0)
    {
        if (!(s = strtok (NULL, "")))
            s = "\1";
        ctcpreply (s_MemoServ, source, "%sVERSION Paradox IRC Services %s%s", s, version_number, s);
    }
    else if (stricmp (cmd, "\1USERINFO\1") == 0)
    {
	if (!(s = strtok (NULL, "")))
	    s = "\1";
	ctcpreply (s_MemoServ, source, "%sUSERINFO 115 107 111 108 100%s", s, s);
    }

    else if (*cmd == '\1')
        return;

    else
    {
	Hash *command, hash_table[] =
	{
	    {"HELP", H_NONE, do_help},
	    {"SEND", H_NONE, do_send},
	    {"OPERSEND", H_IRCOP, do_opersend},
	    {"CSOPSEND", H_CSOP, do_sopsend},
	    {"GLOBALMEMO", H_SRA, do_globalmemo},
	    {"LIST", H_NONE, do_list},
	    {"READ", H_NONE, do_read},
	    {"DEL", H_NONE, do_del},
	    {"OHELP", H_IRCOP, do_ohelp},
	    {NULL}
	};

	if (command = get_hash (source, cmd, hash_table))
	    (*command->process) (source);
	else
	    notice (s_MemoServ, source, ERR_UNKNOWN_COMMAND, strupper (cmd), s_MemoServ);
    }
}

/* Load the MemoServ database into memory. */
void load_ms_dbase (void)
{
    FILE *f = fopen (memoserv_db, "r");
    int i, j;
    MemoList *ml;
    Memo *memos;

    if (!f)
	return;

    switch (i = get_file_version (f, memoserv_db))
    {
    case 6:
    case 5:
    case 4:
    case 3:
    case 2:
    case 1:
	for (i = 33; i < 256; ++i)
	    while (fgetc (f) == 1)
	    {
		ml = smalloc (sizeof (MemoList));
		if (1 != fread (ml, sizeof (MemoList), 1, f))
		    fatal_perror ("[ERR] Read error on %s", memoserv_db);
		alpha_insert_memolist (ml);
		ml->memos = memos = smalloc (sizeof (Memo) * ml->n_memos);
		fread (memos, sizeof (Memo), ml->n_memos, f);
		for (j = 0; j < ml->n_memos; ++j, ++memos)
		    memos->text = read_string (f, memoserv_db);
	    }
	break;
    default:
	fatal ("Unsupported version number (%d) on %s", i, memoserv_db);
    }
    fclose (f);
}

/* Save the MemoServ db in memory to disk. */
void save_ms_dbase (void)
{
    FILE *f;
    int i, j;
    MemoList *ml;
    Memo *memos;
    char memoservsave[2048];

    strcpy (memoservsave, memoserv_db);
    strcat (memoservsave, ".save");
    remove (memoservsave);
    if (rename (memoserv_db, memoservsave) < 0)
	log_perror ("[ERR] Can't back up %s", memoserv_db);
    f = fopen (memoserv_db, "w");
    if (!f)
    {
	log_perror ("[ERR] Can't write to MemoServ database %s", memoserv_db);
	if (rename (memoservsave, memoserv_db) < 0)
	    fatal_perror ("[ERR] Can't restore backup copy of %s", memoserv_db);
	return;
    }

    if (fchown (fileno (f), -1, file_gid) < 0)
	log_perror ("[ERR] Can't set group of %s to %d", memoserv_db, file_gid);

    write_file_version (f, memoserv_db);
    for (i = 33; i < 256; ++i)
    {
	for (ml = memolists[i]; ml; ml = ml->next)
	{
	    fputc (1, f);
	    if (1 != fwrite (ml, sizeof (MemoList), 1, f) ||
		ml->n_memos !=
		fwrite (ml->memos, sizeof (Memo), ml->n_memos, f))
		fatal_perror ("[ERR] Write error on %s", memoserv_db);
	    for (memos = ml->memos, j = 0; j < ml->n_memos; ++memos, ++j)
		write_string (memos->text, f, memoserv_db);
	}
	fputc (0, f);
    }
    fclose (f);
    remove (memoservsave);
}

/* See if the given nick has any waiting memos, and send a NOTICE to
   that nick if so.
 */
void check_memos (const char *nick, const char *checker)
{
    MemoList *ml = find_memolist (nick);

    if (ml)
    {
        int i, newcnt = 0;
        for (i = 0; i < ml->n_memos; i++) 
        {
            if (ml->memos[i].flags & MF_UNREAD)
                newcnt++;
        }

        if (newcnt > 0) 
        {
            notice (s_MemoServ, checker, MS_YOU_HAVE, newcnt,
                    newcnt == 1 ? "" : "s", nick);
            notice (s_MemoServ, checker, MS_LIST_ALL, s_MemoServ);
        }
	/* Warn them of a full or nearly full memobox. This works out to
	   90% full.
	 */
	if (ml->n_memos >= .90 * memobox_limit)
	{
	    if (ml->n_memos >= memobox_limit) /* > since sras bypass it */
		notice (s_MemoServ, checker, MS_FULL_BOX);
	    else
		 notice (s_MemoServ, checker, MS_NEAR_FULL_BOX, ml->n_memos, memobox_limit);
	    notice (s_MemoServ, checker, MS_PLEASE_EMPTY);
	}
    }
}

/* Find the memo list for a given nick.  Return NULL if none. */ 
MemoList *find_memolist (const char *nick)
{
    MemoList *ml;
    int i = -1;
    NickInfo *ni = host (findnick (nick));

    if (ni)
    {
	for (ml = memolists[tolower (*(ni->nick))];
	   ml && (i = stricmp (ml->nick, ni->nick)) < 0;
	     ml = ml->next);
    }
    return i == 0 ? ml : NULL;
}

/* Insert a memo list alphabetically into the db. */
static void alpha_insert_memolist (MemoList * ml)
{
    MemoList *ml2, *ml3;
    char *nick = ml->nick;

    for (ml3 = NULL, ml2 = memolists[tolower (*nick)];
	 ml2 && stricmp (ml2->nick, nick) < 0;
	 ml3 = ml2, ml2 = ml2->next)
	;
    ml->prev = ml3;
    ml->next = ml2;
    if (!ml3)
	memolists[tolower (*nick)] = ml;
    else
	ml3->next = ml;
    if (ml2)
	ml2->prev = ml;
}

/* Remove a nick's memo list from the database. Assumes that the memo
   count for the nick is non-zero.
 */
void del_memolist (MemoList * ml)
{
    if (!ml) return;
    if (ml->next)
	ml->next->prev = ml->prev;
    if (ml->prev)
	ml->prev->next = ml->next;
    else
	memolists[tolower (*ml->nick)] = ml->next;
    if (ml->memos)
	free (ml->memos);
    free (ml);
}

/* Return a help message. */
static void do_help (const char *source)
{
    char *cmd = strtok (NULL, "");
    char buf[BUFSIZE];

    if (!cmd)
    {
	notice_list (s_MemoServ, source, ms_help_start);
	notice_list (s_MemoServ, source, ms_help);
	notice_list (s_MemoServ, source, ms_help_end);
    }
    else
    {
	Hash_HELP *command, hash_table[] =
	{
	    {"DEL", H_NONE, del_help},
	    {"READ", H_NONE, read_help},
	    {"SEND", H_NONE, send_help},
	    {"LIST", H_NONE, list_help},
	    {NULL}
	};
 
	if (command = get_help_hash (source, cmd, hash_table))
	{
	    notice_list (s_MemoServ, source, ms_help_start);
	    notice_list (s_MemoServ, source, command->process);
	    notice_list (s_MemoServ, source, ms_help_end);
	}
	else
	    notice (s_MemoServ, source, ERR_NO_HELP, cmd);
    }
}

/* Return an OperHelp message. */ 
static void do_ohelp (const char *source)
{
    char *cmd = strtok (NULL, "");
    char buf[BUFSIZE];

    if (!cmd)
    {
        notice_list (s_MemoServ, source, ms_ohelp_start);
	notice_list (s_MemoServ, source, ms_ohelp);
        if (is_services_op (source))
            notice_list (s_MemoServ, source, ms_sop_help);
	if (is_services_admin (source))
	    notice_list (s_MemoServ, source, ms_admin_help);
        notice_list (s_MemoServ, source, ms_ohelp_end);
    }
    else
    {
	Hash_HELP *command, hash_table[] =
	{
	    {"OPERSEND", H_IRCOP, opersend_help},
	    {"CSOPSEND", H_CSOP, sopsend_help},
	    {"GLOBALMEMO", H_SRA, globalmemo_help},
	    {NULL}
	};

	if (command = get_help_hash (source, cmd, hash_table))
	{
	    notice_list (s_MemoServ, source, ms_ohelp_start);
	    notice_list (s_MemoServ, source, command->process);
	    notice_list (s_MemoServ, source, ms_ohelp_end);
	}
        else 
	    notice (s_MemoServ, source, ERR_NO_HELP, cmd);
    }
}

/* Are we clear to send a memo? If so, return 1. If for some reason
   we can't send a memo, return 0. 
 */
static int candomemo (const char *source, const char *other)
{
    NickInfo *ni, *hni;

    if (!(ni = findnick (source)) || !(hni = host (ni)))
	notice (s_MemoServ, source, NS_YOU_NOT_REGISTERED, source);

    else if (!userisnick (source))
	notice (s_MemoServ, source, ERR_IDENT_NICK_FIRST, s_NickServ);

    else if (other)
	if (!findnick (other))
	    notice (s_MemoServ, source, NS_NOT_REGISTERED, other);
	else if ((findnick (other)->flags & NI_VERBOTEN))
	    notice (s_MemoServ, source, NS_FORBIDDEN, other);
	else if ((findnick (other)->flags & NI_SUSPENDED))
	    notice (s_MemoServ, source, NS_SUSPENDED, other);
	else if (is_on_ignore (source, other) && !(hni->flags & NI_IRCOP))
	    notice (s_MemoServ, source, NS_AM_IGNORED, other);
        else if (host (findnick (other))->flags & NI_NOMEMO && !is_justservices_admin (source))
                notice (s_MemoServ, source, MS_NOMEMO, other);
	else
	    return 1;

    else
	return 1;

    return 0;
}

/* Send a memo to a nick. */
static void do_send (const char *source)
{
    User *u = finduser (source);
    Memo *m;
    char *arg = strtok (NULL, " ");
    char *text = strtok (NULL, "");
    unsigned int len;

    if (!text)
    {
	notice (s_MemoServ, source, "Syntax: \2SEND Nick Text\2");
	notice (s_MemoServ, source, ERR_MORE_INFO, s_MemoServ, "SEND");
	return;
    }

    if (u)
	if (!is_oper (source))
	    if (time (NULL) - u->lastmemo < memo_delay)
	    {
		notice (s_MemoServ, source, ERR_WAIT_SEND, disect_time ((u->lastmemo - time (NULL) + memo_delay), 0));
		return;
	    }

    if (candomemo (source, arg))
    {
	int i;
	NickInfo *ni = findnick (source);
	NickInfo *hni = findnick (arg);
	MemoList *ml = find_memolist (hni->nick);

	if (!host (ni)->flags & NI_IDENTIFIED)
	{
	    notice (s_MemoServ, source, ERR_NEED_PASSWORD, "SEND");
	    notice (s_MemoServ, source, ERR_IDENT_NICK_FIRST, s_NickServ);
	    return;
	}

	if ((len = strlen(text)) > 256)
	{
	    notice (s_MemoServ, source, ERR_MEMO_TOO_LONG, len, len - 256,
			len - 256 > 1 ? "s" : "");
	    return;
	}

	if (ml && ml->n_memos > memobox_limit - 1) 
	{
	    notice (s_MemoServ, source, ERR_FULL_MEMOBOX, arg);
	    return;
	}

	if (!ml)
	{
	    ml = scalloc (sizeof (MemoList), 1);
	    strscpy (ml->nick, hni->nick, NICKMAX);
	    alpha_insert_memolist (ml);
	}

	++ml->n_memos;
	u->lastmemo = time (NULL);
	ml->memos = srealloc (ml->memos, sizeof (Memo) * ml->n_memos);
	m = &ml->memos[ml->n_memos - 1];
	strscpy (m->sender, source, NICKMAX);

	if (ml->n_memos > 1)
	{
	    m->number = m[-1].number + 1;
	    if (m->number < 1)
	    {
		int i;
		for (i = 0; i < ml->n_memos; ++i)
		    ml->memos[i].number = i + 1;
	    }
	}
	else
	    ml->memos[ml->n_memos - 1].number = 1;
	m->time = time (NULL);
	m->text = sstrdup (text);
	m->flags = MF_UNREAD;
	notice (s_MemoServ, source, MS_SEND, hni->nick);
	if (userisnick (hni->nick))
	{
	    notice (s_MemoServ, hni->nick, MS_NEW, "", source);
	    notice (s_MemoServ, hni->nick, MS_READ_NEW, s_MemoServ, m->number);
	    /* Warn them of a full or nearly full memobox. This works out to
	       90% full.
	     */
	    if (ml->n_memos >= .90 * memobox_limit)
	    {
		if (ml->n_memos >= memobox_limit)
		    notice (s_MemoServ, hni->nick, MS_FULL_BOX);   
		else
		    notice (s_MemoServ, hni->nick, MS_NEAR_FULL_BOX, ml->n_memos, memobox_limit);
		notice (s_MemoServ, hni->nick, MS_PLEASE_EMPTY);
	    }
	}
	i = slavecount (hni->nick);
	for (i = slavecount (hni->nick); i; i--)
	    if (userisnick (slave (hni->nick, i)->nick))
	    {
		notice (s_MemoServ, slave (hni->nick, i)->nick, MS_NEW, "", source);
		notice (s_MemoServ, slave (hni->nick, i)->nick, MS_READ_NEW, s_MemoServ, m->number);
		/* Warn them of a full or nearly full memobox. This works out to
		   90% full.
		 */
		if (ml->n_memos >= .90 * memobox_limit)
		{
		    if (ml->n_memos >= memobox_limit)
			notice (s_MemoServ, slave (hni->nick, i)->nick, MS_FULL_BOX);   
		    else
			notice (s_MemoServ, slave (hni->nick, i)->nick, MS_NEAR_FULL_BOX, ml->n_memos, memobox_limit);
		    notice (s_MemoServ, slave (hni->nick, i)->nick, MS_PLEASE_EMPTY);
		}
	    }

	if ((hni->flags & NI_MEMOMAIL) && hni->email && (memomail_on==TRUE))
	{
	    FILE *fp;
	    char timebuf[64];
	    struct tm tm;
	    char cmdbuf[500];
	    char cmdbuf2[500];
	    int i;
	    int x = 0;

	    if (sendmail_path[0] == '\0')
		return;

	    tm = *localtime(&m->time);
	    strftime (timebuf, sizeof (timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);	
	    timebuf[sizeof(timebuf)-1] = 0;

	    m->flags &= ~MF_UNREAD;
	    m->flags = MF_EMAILED;

	    for (i = 0, m = ml->memos; i < ml->n_memos; ++i, ++m)
	    {
		if (m->flags & MF_EMAILED)
		{
		    ++x;
		    if (x > 5)
		    { 
			/* More than 5 emailed memos found, abort */
			m->flags &= ~MF_EMAILED;
			m->flags = MF_UNREAD;
			return;
		    }
		}
	    }
	    sprintf (cmdbuf, "%s %s", sendmail_path, hni->email);
	    fp = popen (cmdbuf, "w");
	    fprintf (fp, "Date: %s\n", timebuf);
	    fprintf (fp, "From: %s <%s>\n", s_MemoServ, ms_email);
	    fprintf (fp, "To: %s <%s>\n", hni->nick, hni->email);
	    fprintf (fp, "Subject: New memo from %s\n", source);
	    fprintf (fp, "X-Abuse: %s\n\n", abuse_email);
	    fprintf (fp, "You have a new memo from %s. Here is the memo:\n\n", source);
	    fprintf (fp, "%s\n\n", text);
	    fprintf (fp, "This message was generated by the %s %s.\n", network_name, s_MemoServ);
	    if (contact_url[0] != '\0')
	    {
		fprintf (fp, "If you have and problems or questions about this service, please view:\n");
		fprintf (fp, "%s\n", contact_url);
	    }
	    pclose (fp);
	}
    }
}

/* Send a memo to all nicks with the IRCOp flag on. 
   This does NOT email the memos. 
 */
static void do_opersend (const char *source)
{
    char *text = strtok (NULL, "");

    if (!text)
    {
	notice (s_MemoServ, source, "Syntax: \2OPERSEND Text\2");
	notice (s_MemoServ, source, ERR_MORE_INFO, s_MemoServ, "OPERSEND");
    }
    else
    {
	if (candomemo (source, NULL))
	{
	    int i, j;
	    MemoList *ml;
	    Memo *m;
	    NickInfo *hni, *ni = host (findnick (source));
	    char sender[NICKMAX];
	    char *text2;
	    text2=smalloc(strlen(text) + 16);
	    if (!ni)
	    {
		notice (s_MemoServ, source, NS_YOU_NOT_REGISTERED, source);
		free(text2);
		return;
	    }
	    else if (!(ni->flags & NI_IRCOP))
	    {
		notice (s_MemoServ, source, ERR_MUST_BE_IRCOP, "OPERSEND");
		free(text2);
		return;
	    }
	    strscpy (sender, ni->nick, NICKMAX);
	    for (i = 33; i < 256; ++i)
	    {
		for (ni = nicklists[i]; ni; ni = ni->next)
		{
		    if ((ni->flags & NI_IRCOP) && stricmp (ni->nick, sender) != 0)
		    {
			ml = find_memolist (ni->nick);
			if (!ml)
			{
			    ml = scalloc (sizeof (MemoList), 1);
			    strscpy (ml->nick, ni->nick, NICKMAX);
			    alpha_insert_memolist (ml);
			}
			++ml->n_memos;
			ml->memos = srealloc (ml->memos, sizeof (Memo) * ml->n_memos);
			m = &ml->memos[ml->n_memos - 1];
			strscpy (m->sender, source, NICKMAX);
			if (ml->n_memos > 1)
			{
			    m->number = m[-1].number + 1;
			    if (m->number < 1)
			    {
				for (j = 0; j < ml->n_memos; ++j)
				    ml->memos[j].number = j + 1;
			    }
			}
			else
			    ml->memos[ml->n_memos - 1].number = 1;
			snprintf (text2, strlen(text) + 16, "[\2OPER\2] %s", text);
			m->time = time (NULL);
			m->text = sstrdup (text2);
			m->flags = MF_UNREAD;
			hni = host (ni);
			if (userisnick (hni->nick))
			{
			    notice (s_MemoServ, hni->nick, MS_NEW, "oper", source);
			    notice (s_MemoServ, hni->nick, MS_READ_NEW, s_MemoServ, m->number);
			}
			for (j = slavecount (hni->nick); j; j--)
			{
			    if (userisnick (slave (hni->nick, j)->nick))
			    {
				notice (s_MemoServ, slave (hni->nick, j)->nick, MS_NEW, "oper", source);
				notice (s_MemoServ, slave (hni->nick, j)->nick, MS_READ_NEW, s_MemoServ, m->number);
			    }
			}
		    }
		}
	    }
	    notice (s_MemoServ, source, MS_MASS_SEND, "IRCOps");
	    free(text2);
	}
    }
}

/* Send a memo to all nicks with the CSOp flag on.
   This does NOT email the memos. 
 */
static void do_sopsend (const char *source)
{
    char *text = strtok (NULL, "");

    if (!text)
    {
	notice (s_MemoServ, source, "Syntax: \2CSOPSEND Text\2");
	notice (s_MemoServ, source, ERR_MORE_INFO, s_MemoServ, "CSOPSEND");
    }
    else
    {
	if (candomemo (source, NULL))
	{
	    int i, j;
	    MemoList *ml;
	    Memo *m;
	    NickInfo *hni, *ni = host (findnick (source));
	    char sender[NICKMAX];
	    char *text2;
	    text2=smalloc(strlen (text) + 16);
	    snprintf (text2, strlen (text) + 16, "[\2CSOp\2] %s", text);
	    if (!ni)
	    {
		notice (s_MemoServ, source, NS_YOU_NOT_REGISTERED, source);
		free(text2);
		return;
	    }
	    strscpy (sender, ni->nick, NICKMAX);
	    for (i = 0; i < nsop; ++i)
	    {
		if (stricmp (sender, sops[i].nick) != 0)
		{
		    ml = find_memolist (sops[i].nick);
		    if (!ml)
		    {
			ml = scalloc (sizeof (MemoList), 1);
			strscpy (ml->nick, sops[i].nick, NICKMAX);
			alpha_insert_memolist (ml);
		    }
		    ++ml->n_memos;
		    ml->memos = srealloc (ml->memos, sizeof (Memo) * ml->n_memos);
		    m = &ml->memos[ml->n_memos - 1];
		    strscpy (m->sender, source, NICKMAX);
		    if (ml->n_memos > 1)
		    {
			m->number = m[-1].number + 1;
			if (m->number < 1)
			{
			    for (j = 0; j < ml->n_memos; ++j)
				ml->memos[j].number = j + 1;
			}
		    }
		    else
			ml->memos[ml->n_memos - 1].number = 1;
		    m->time = time (NULL);
		    m->text = sstrdup (text2);
		    m->flags = MF_UNREAD;
		    hni = host (findnick (sops[i].nick));
		    if (userisnick (hni->nick))
		    {
			notice (s_MemoServ, hni->nick, MS_NEW, "CSOp", source);
			notice (s_MemoServ, hni->nick, MS_READ_NEW, s_MemoServ, m->number);
		    }
		    for (j = slavecount (hni->nick); j; j--)
		    {
			if (userisnick (slave (hni->nick, j)->nick))
			{
			    notice (s_MemoServ, slave (hni->nick, j)->nick, MS_NEW, "CSOp", source);
			    notice (s_MemoServ, slave (hni->nick, j)->nick, MS_READ_NEW, s_MemoServ, m->number);
			}
		    }
		}
	    }
	    notice (s_MemoServ, source, MS_MASS_SEND, "CSOps");
	    free(text2);
	}
    }
}

/* List the memos (if any) for the source nick. */
static void do_list (const char *source)
{
    Memo *m;
    int i;
    char *arg = strtok (NULL, "");
    char listbuf[BUFSIZE];

    if (candomemo (source, NULL))
    {
	NickInfo *hni = host (findnick (source));
	MemoList *ml = find_memolist (hni->nick);

	if (!ml)
	    notice (s_MemoServ, source, MS_YOU_DONT_HAVE);

	else
	{
	    notice (s_MemoServ, source, MS_LIST, s_MemoServ);
	    notice (s_MemoServ, source, "Num  Status  Sender            Waiting for");

	    for (i = 0, m = ml->memos; i < ml->n_memos; ++i, ++m)
	    {

		*listbuf = 0;

                if (m->flags & MF_EMAILED)
		    strcat (listbuf, "Mailed");
		else if (m->flags & MF_UNREAD)
		    strcat (listbuf, "Unread");
		else if (!*listbuf)
		    strcat (listbuf, "Read");

		notice (s_MemoServ, source, "%-4d %-7s %-16s  %s", m->number, listbuf, m->sender, time_ago (m->time, 1));

	    }
	}
    }
}

/* Read a memo. */
static void do_read (const char *source)
{
    Memo *m;
    char numstr[8];
    int num;
    char *arg = strtok (NULL, " ");

    if (arg)
	strscpy (numstr, arg, sizeof(numstr));

    if (!arg)
    {
	notice (s_MemoServ, source, "Syntax: \2READ Num|All");
	notice (s_MemoServ, source, ERR_MORE_INFO, s_MemoServ, "READ");
    }

    else if (!numstr || ((num = atoi (numstr)) <= 0 && stricmp (numstr, "ALL") != 0))
    {
	notice (s_MemoServ, source, "Syntax: \2READ Num|All");
	notice (s_MemoServ, source, ERR_MORE_INFO, s_MemoServ, "READ");
    }

    else
    {
	if (candomemo (source, NULL))
	{
	    NickInfo *ni = findnick (source);
	    MemoList *ml = find_memolist (host (ni)->nick);

	    if (!host (ni)->flags & NI_IDENTIFIED)
	    {
		notice (s_MemoServ, source, ERR_NEED_PASSWORD, "READ");
		notice (s_MemoServ, source, ERR_IDENT_NICK_FIRST, s_NickServ);
		return;
	    }

	    if (!ml)
	        notice (s_MemoServ, source, MS_YOU_DONT_HAVE);

	    else if (num > 0)
	    {
		int i;
		for (i = 0; i < ml->n_memos; ++i)
		if (ml->memos[i].number == num)
		    break;
		if (i >= ml->n_memos)
		    notice (s_MemoServ, source, MS_DOESNT_EXIST, num);
		else
		{
		    m = &ml->memos[i];
		    notice (s_MemoServ, source, MS_MEMO,
			m->number, m->sender, time_ago (m->time, 1));
		    notice (s_MemoServ, source, "%s", m->text);
		    m->flags &= ~MF_UNREAD;
		    m->flags &= ~MF_EMAILED;
		    notice (s_MemoServ, source, MS_TODEL,
			s_MemoServ, m->number);
		}
	    }
	    else
	    {
		int i;
		for (i = 0; i < ml->n_memos; ++i)
		{
		    m = &ml->memos[i];
		    notice (s_MemoServ, source, MS_MEMO,
			m->number, m->sender, time_ago (m->time, 1));
		    notice (s_MemoServ, source, "%s", m->text);
		    m->flags &= ~MF_UNREAD;
		}
		    notice (s_MemoServ, source, MS_TODEL_ALL, s_MemoServ);
	    }
	}
    }
}

/* Delete a memo. */
static void do_del (const char *source)
{
    char *arg = strtok (NULL, " ");
    char *arg2 = strtok (NULL, "");
    char numstr[8];
    int num, i;

    if (arg)
    {
	strscpy (numstr, arg, sizeof(numstr));
    }

    if (!arg)
    {
	notice (s_MemoServ, source, "Syntax: \2DEL Num|All");
	notice (s_MemoServ, source, ERR_MORE_INFO, s_MemoServ, "DEL");
    }
    else if (!numstr ||
	     ((num = atoi (numstr)) <= 0 && stricmp (numstr, "ALL") != 0))
    {
	notice (s_MemoServ, source, "Syntax: \2DEL Num|All");
	notice (s_MemoServ, source, ERR_MORE_INFO, s_MemoServ, "DEL");
    }

    else
    {
	if (candomemo (source, NULL))
	{
	    NickInfo *hni = host(findnick(source));
	    MemoList *ml = find_memolist (hni->nick);
	    if (!ml)
		notice (s_MemoServ, source, MS_YOU_DONT_HAVE);
	    else
	    {
		if (num > 0)
		{
		    /* Delete a specific memo. */
		    for (i = 0; i < ml->n_memos; ++i)
			if (ml->memos[i].number == num)
			    break;
		    if (i < ml->n_memos)
		    {
			if (ml->memos[i].text)
			    free (ml->memos[i].text);
			--ml->n_memos;
			if (i < ml->n_memos)
			    bcopy (ml->memos + i + 1, ml->memos + i,
				   sizeof (Memo) * (ml->n_memos - i));
			notice (s_MemoServ, source, MS_DELETE, num);
		    }
		    else
			notice (s_MemoServ, source, MS_DOESNT_EXIST, num);
		}
		else
		{
		    /* Delete all memos.  This requires freeing the memory
		     * holding the text of each memo and flagging that there
		     * are no more left. */
		    for (i = 0; i < ml->n_memos; ++i)
			if (ml->memos[i].text)
			    free (ml->memos[i].text);
		    ml->n_memos = 0;
		    notice (s_MemoServ, source, MS_DELETE_ALL);
		}
		/* Did we delete the last memo?  If so, delete this MemoList. */
		if (ml->n_memos == 0)
		    del_memolist (ml);
	    }
	}
    }
}

/* Send a memo to EVERY registered nick. This is basically copy/paste
   of opersend, but without the checks for the ircop flag. 
 */
static void do_globalmemo (const char *source)
{
    char *text = strtok (NULL, "");
            
    if (!text)       
    {
	notice (s_MemoServ, source, "Syntax: \2GLOBALMEMO Text\2");
	notice (s_MemoServ, source, ERR_MORE_INFO, s_MemoServ, "GLOBALMEMO");
    }

    else
    {
	if (candomemo (source, NULL))
	{   
	    int i, j;
	    MemoList *ml;
	    Memo *m;
	    NickInfo *hni, *ni = host (findnick (source));
	    char sender[NICKMAX];
	    char *text2;
	    text2=smalloc(strlen(text) + 16);

	    /* This really shouldn't fail.. */
	    if (!ni)
	    {
		notice (s_MemoServ, source, NS_YOU_NOT_REGISTERED, source);
		free(text2);
		return;
	    }

	    strscpy (sender, ni->nick, NICKMAX);

	    for (i = 33; i < 256; ++i)
	    {
		for (ni = nicklists[i]; ni; ni = ni->next)
		{
		    /* Only the hosts gets a copy */
		    if (!(ni->flags & NI_SLAVE))
		    {
			ml = find_memolist (ni->nick);

			if (!ml)
			{
			    ml = scalloc (sizeof (MemoList), 1);
			    strscpy (ml->nick, ni->nick, NICKMAX);
			    alpha_insert_memolist (ml);
			}

			++ml->n_memos;
			ml->memos = srealloc (ml->memos, sizeof (Memo) * ml->n_memos);
			m = &ml->memos[ml->n_memos - 1];
			/* We store the sender's nickname as the network's
			   name if defined, else as MemoServ's nick.
			 */
			strscpy (m->sender, network_name ? network_name : s_MemoServ, NICKMAX);

			if (ml->n_memos > 1)
			{
			    m->number = m[-1].number + 1;
			    if (m->number < 1)
			    {
				for (j = 0; j < ml->n_memos; ++j)  
				    ml->memos[j].number = j + 1;
			    }
			}

			else
			    ml->memos[ml->n_memos - 1].number = 1;
			snprintf (text2, strlen(text) + 16, "[\2Global\2] %s", text);
			m->time = time (NULL);
			m->text = sstrdup (text2);
			m->flags = MF_UNREAD;
			hni = host (ni);
			if (userisnick (hni->nick))
			{
			    notice (s_MemoServ, hni->nick, MS_NEW_GLOBAL);
			    notice (s_MemoServ, hni->nick, MS_READ_NEW, s_MemoServ, m->number);
			}
			for (j = slavecount (hni->nick); j; j--)
			{
			    if (userisnick (slave (hni->nick, j)->nick))
			    {
				notice (s_MemoServ, slave (hni->nick, j)->nick, MS_NEW_GLOBAL);
				notice (s_MemoServ, slave (hni->nick, j)->nick, MS_READ_NEW, s_MemoServ, m->number);
			    }
			}
		    }
		}
	    }
	    notice (s_MemoServ, source, MS_GLOBAL_SEND);
	    wallops (s_MemoServ, MS_GLOBAL_WALLOP, source);
	    free(text2);
	}
    }
}

