/* Main processing code for Services.
  
   Paradox IRC Services - (c) 2001 In Mean
       E-Mail: <paradox@auspice.org> IRC: auspice@Auspice.org
   WreckedNet IRC Services - (c) 2000 Darcy Grexton
       E-Mail: <skold@wrecked.net> IRC: skold@WreckedNet
   Based on Magick 1.4
  
   Magick IRC Services are copyright (c) 1996-1998 Preston A. Elder.
       E-mail: <prez@magick.tm>   IRC: PreZ@RelicNet
   Originally based on EsperNet services (c) 1996-1998 Andy Church
   This program is free but copyrighted software; see the file COPYING for
   details.
 */

#include "services.h"

/* People to ignore (hashed by first character of nick). */
int ignorecnt = 0, ignore_size = 0;
Ignore *ignore = NULL;
int serv_size = 0;
Servers *servlist = NULL;
Timer *pings = NULL;
int flood_messages;
int flood_time;
int ignore_time;
int ignore_offences;
time_t last_squit;

static void add_msg (const char *nick);
static void trigger_flood (User * u);

void addtimer (const char *label)
{
    Timer *p;

    if (!gettimer (label))
    {
	p = smalloc (sizeof (Timer));
	p->next = pings;
	p->prev = NULL;
	if (pings)
	    pings->prev = p;
	pings = p;
	p->label = sstrdup (label);
	p->start = time (NULL);
    }
}

void deltimer (const char *label)
{
    Timer *p;

    for (p = pings; p && stricmp (p->label, label) != 0; p = p->next);
	if (p)
	{
	    if (p->next)
		p->next->prev = p->prev;
	    if (p->prev)
		p->prev->next = p->next;
	    else
		pings = p->next;
	    if (p->label)
		free (p->label);
	    free (p);
	}
}

time_t gettimer (const char *label)
{
    Timer *p;

    for (p = pings; p && stricmp (p->label, label) != 0; p = p->next);
	if (p)
	    return p->start;
    return 0;
}

static void add_msg (const char *nick)
{
    User *u = finduser (nick);
    int i;
    if (is_ignored (nick))
	return;
    if (is_oper (nick))
	return;
    if (u)
    {
	if (u->messages == flood_messages)
	    u->messages--;
	for (i = 0; i < flood_messages; i++)
	    if (i + 1 < flood_messages && i + 1 < u->messages)
		u->msg_times[i] = u->msg_times[i + 1];
	    else
	    {
		u->msg_times[i] = time (NULL);
		u->messages++;
		break;
	    }
	if (u->messages == flood_messages &&
	    (time (NULL) - u->msg_times[0] < flood_time))
	    trigger_flood (u);
    }
}

/* Too many messages in too little time */
static void trigger_flood (User * u)
{
    u->messages = 0;
    u->flood_offence++;
    if (ignorecnt >= ignore_size)
    {
	if (ignore_size < 8)
	    ignore_size = 8;
	else
	    ignore_size *= 2;
	ignore = srealloc (ignore, sizeof (*ignore) * ignore_size);
    }
    strscpy (ignore[ignorecnt].nick, u->nick, NICKMAX);
    if (u->flood_offence < ignore_offences)
    {
	wallops (s_OperServ, FLOODING, u->nick, "TEMP");
	notice (s_OperServ, u->nick, TEMP_FLOOD,
		flood_messages, flood_time, ignore_time);
	ignore[ignorecnt].start = time (NULL);
    }
    else
    {
	wallops (s_OperServ, FLOODING, u->nick, "PERM");
	notice (s_OperServ, u->nick, PERM_FLOOD,
		ignore_offences);
	u->flood_offence = 0;
	ignore[ignorecnt].start = 0;
    }
    ignorecnt++;
}

int is_ignored (const char *nick)
{
    int i;
    if (is_oper (nick))
	return 0;
    for (i = 0; i < ignorecnt; ++i)
	if (stricmp (nick, ignore[i].nick) == 0)
	    return 1;
    return 0;
}

/* split_buf:  Split a buffer into arguments and store the arguments in an
               argument vector pointed to by argv (which will be malloc'd
               as necessary); return the argument count.  All argument
               values will also be malloc'd.  If colon_special is non-zero,
               then treat a parameter with a leading ':' as the last
               parameter of the line, per the IRC RFC. Destroys the buffer
               by side effect.
 */

int split_buf (char *buf, char ***argv, int colon_special)
{   
    int argvsize = 8;
    int argc;
    char *s;
 
    *argv = smalloc(sizeof(char *) * argvsize);
    argc = 0;
    while (*buf)
    {
	if (argc == argvsize)
	{
	    argvsize += 8;
	    *argv = srealloc(*argv, sizeof(char *) * argvsize);
	}
	if (*buf == ':')
	{
	    (*argv)[argc++] = buf+1;
	    buf = "";
	}
	else
	{
	    s = strpbrk(buf, " ");
	    if (s)
	    {
		*s++ = 0;
		while (isspace(*s))
		    s++;
	    }
	    else
		s = buf + strlen(buf);
	    (*argv)[argc++] = buf;
	    buf = s;
	}
    }
    return argc;
}

/* process:  Main processing routine.  Takes the string in inbuf (global
             variable) and does something appropriate with it.
 */
void process ()
{
    char source[64];
    char cmd[64];
    char buf[BUFSIZE];
    char *s;
    int ac, i;
    char **av;

    /* If debugging, log the buffer. */
    if (runflags & RUN_DEBUG)
	debug ("--> %s", inbuf);

    /* First make a copy of the buffer so we have the original in case we
       crash - in that case, we want to know what we crashed on.
     */
    strscpy (buf, inbuf, sizeof (buf));

    /* Split the buffer into pieces. */
    if (*buf == ':')
    {
	s = strpbrk (buf, " ");
	if (!s)
	    return;
	*s = 0;
	while (isspace (*++s))
	    ;
	strscpy (source, buf+1, sizeof(source));
	strscpy (buf, s, BUFSIZE);
    }
    else
	*source = 0;
    if (!*buf)
	return;
    s = strpbrk (buf, " ");
    if (s)
    {
	*s = 0;
	while (isspace (*++s))
	    ;
    }
    else
	s = buf + strlen (buf);
    strscpy (cmd, buf, sizeof (cmd));
    ac = split_buf (s, &av, 1);

    /* Do something with the command. */
    if (!strcmp (cmd, "PING") || !strcmp (cmd, "8"))
	send_cmd (server_name, "PONG %s %s", ac > 1 ? av[1] : server_name, av[0]);

    else if (!strcmp (cmd, "PONG") || !strcmp (cmd, "9"))
    {
	int i;
	for (i = 0; i < servcnt; i++)
	    if (stricmp (servlist[i].server, av[0]) == 0)
		servlist[i].lag = time (NULL) - gettimer(av[0]);
	deltimer(av[0]);
    }
    else if (!strcmp (cmd, "436"))
    {
	User *u;
	introduce_user (av[0]);
	if ((u = finduser (av[0])))
	    delete_user (u);
    }
#ifdef IRC_UNREAL
    else if (!strcmp (cmd, "302"))
    {
        User *u;
        NickInfo *ni, *hni;
        char *nick = strtok(av[1],"=");
        char *virthost;

        if (!nick)
                return;

        if (strchr(nick, '*')) {
                virthost = strtok(NULL,"@");
                virthost = strtok(NULL," ");
                nick = strtok(nick, "*");
        } else {
                virthost = strtok(NULL,"@");
                virthost = strtok(NULL," ");
        }

        if (!virthost)
                return;

        u = finduser (nick);

        if (!u)
                return;

        if (u && virthost)
                u->host = sstrdup(virthost);

    }
#endif
    else if (!strcmp (cmd, "401"))
    {
	User *u;
	if ((u = finduser (av[1])))
	    delete_user (u);
    }
    else if (!strcmp (cmd, "AWAY") || !strcmp (cmd, "6"))
    {
	int i;
	if (ac == 0 || *av[0] == 0)
	    if(memoserv_on==TRUE)
		check_memos (source, source);
    }
    else if (!strcmp (cmd, "351"))
    {
        for (i = 0; i < servcnt && stricmp (servlist[i].server, source) != 0; i++);
	{
	    if (i < servcnt)
	    {
		servlist[i].version = sstrdup (av[1]);
		servlist[i].versionflags = sstrdup (av[3]);
	    }
	}
    }
#ifdef IRC_UNREAL
    else if (!strcmp (cmd, "NETINFO") || !strcmp (cmd, "AO"))
    {
	send_cmd (server_name, "NETINFO 0 %lu %s * * * * :%s",
                time(NULL), unreal_protocol ? unreal_protocol : "0", av[7]);
    }
    else if (!strcmp (cmd, "SETNAME") || !strcmp (cmd, "AE"))
    {
        User *u;
        NickInfo *ni;
            
        /* First, update the user structure. */
        u = finduser (source);
        if (u)
            u->realname = sstrdup(av[0]);
        /* Now, update the nickinfo. */
        ni = findnick (u->nick);
        if (ni) 
        {
            if (ni->last_realname)
                free (ni->last_realname);
            ni->last_realname = sstrdup (av[0]);
        }
    }
    else if (!strcmp (cmd, "UMODE2") || !strcmp (cmd, "|"))
    {
        av[1] = sstrdup (av[0]);
        strcpy (av[0], source);
        do_umode (source, ac, av);
    }
    else if (!strcmp (cmd, "SETIDENT") || !strcmp (cmd, "AD"))
    {
        User *u;
        NickInfo *ni;
        /* First, update the user structure. */
        u = finduser (source);  
        if (u)
            u->username = sstrdup(av[0]);
     
        /* Now, update the nickinfo. */
        ni = findnick (u->nick);
	if (ni) 
	{
	    free (ni->last_usermask);
	    ni->last_usermask = smalloc (strlen (u->nick) + strlen (u->username) + strlen (u->host) + 3);
	    snprintf (ni->last_usermask, strlen (u->nick) + strlen (u->username) + strlen (u->host) + 3, "%s@%s", u->username, u->host);
	}
    }
#endif /* IRC_UNREAL */

    else if (!strcmp (cmd, "JOIN") || !strcmp (cmd, "C"))
    {
	if (ac != 1)
	    goto endfunc;
	do_join (source, ac, av);
    }
#ifdef IRC_BAHAMUT
    else if (!strcmp (cmd, "SJOIN"))
	do_sjoin (source, ac, av);
#endif 
    else if (!strcmp (cmd, "KICK") || !strcmp (cmd, "H"))
    {
	if (ac != 3)
	    goto endfunc;
	do_kick (source, ac, av);
    }
    else if (!strcmp (cmd, "KILL") || !strcmp (cmd, "SVSKILL")
	|| !strcmp (cmd, ".") || !strcmp (cmd, "h"))
    {
	int i;
	ChannelInfo *ci;
	if (ac != 2)
	    goto endfunc;
	do_kill (source, ac, av);
	if (is_services_nick (av[0]))
	    introduce_user (av[0]);
	if(chanserv_on==TRUE)
	{
	    if (stricmp (av[0], s_ChanServ) == 0)
		for (i = 33; i < 256; ++i)
		{
		    ci = chanlists[i];
		    while (ci)
		    {
			if (findchan (source))
			    if (ci->flags & CI_JOIN)
				do_cs_join (ci->name);
			    else if (ci->mlock_key || hasmode(ci->mlock_on, "i"))
				do_cs_protect (ci->name);
			ci = ci->next;
		    }
		}
	}
    }
    else if (!strcmp (cmd, "MODE") || !strcmp (cmd, "SVSMODE")
	|| !strcmp (cmd, "G") || !strcmp (cmd, "n"))
    {
	if (validchan(av[0]))
	{
	    if (ac < 2)
		goto endfunc;
	    do_cmode (source, ac, av);
	}
	else
	{
	    if (ac != 2)
		goto endfunc;
	    if (!strcmp (cmd, "SVSMODE") || !strcmp (cmd, "n"))
		do_svumode (source, ac, av);
	    else
		do_umode (source, ac, av);
	}
    }
    else if (!strcmp (cmd, "MOTD") || !strcmp (cmd, "F"))
    {
        FILE *f;
        char buf[BUFSIZE];

	send_cmd (server_name, "375 %s :- %s Message of the Day -", source, server_name);
	send_cmd (server_name, "372 %s :- Paradox IRC Services v%s (c) 2001 In Mean", source, version_number); 

        f = fopen (MOTD_FILE, "r");
        if (f)
        {
            while (fgets (buf, sizeof (buf), f))
            {
                buf[strlen (buf) - 1] = 0;
                send_cmd (server_name, "372 %s :- %s", source, buf);
            }
            do_motd (source);
            fclose (f);
        }
        else
	{
            do_motd (source);
	}
    }
    else if (!strcmp (cmd, "ADMIN") || !strcmp (cmd, "@")) 
    {
        send_cmd (server_name, "256 %s :Administrative info about %s", source, server_name);
        send_cmd (server_name, "257 %s :Paradox IRC Services v%s (c) 2001 In Mean", source, version_number);
        send_cmd (server_name, "258 %s :SRAs: %s", source, services_admin);
        send_cmd (server_name, "259 %s :For more information E-Mail paradox@auspice.org.", source);
    }
    else if (!strcmp (cmd, "USERS") || !strcmp (cmd, "3"))
    {
	send_cmd (server_name, "446 %s :USERS has been disabled", source);
    }
    else if (!strcmp(cmd, "INFO") || !strcmp (cmd, "/"))
    {
	char **text = infotext;

	send_cmd (server_name, "373 %s :Paradox IRC Services v%s --", source, version_number);
	while (*text) {
	send_cmd (server_name, "371 %s :%s", source, *text++); }
	send_cmd (server_name, "371 %s :paradox-services-%s", source, version_number);
	send_cmd (server_name, "371 %s :Birth Date: %s, compile # %s", source, compile_time, version_build);
	send_cmd (server_name, "371 %s :On-line since %s", source, ctime (&start_time));
	send_cmd (server_name, "374 %s :End of /INFO list.", source);
    }
    else if (!strcmp (cmd, "NICK") || !strcmp (cmd, "SVSNICK")
	|| !strcmp (cmd, "&") || !strcmp (cmd, "e"))
	do_nick (source, ac, av);
    else if (!strcmp (cmd, "PART") || !strcmp (cmd, "D"))
    {
	if (ac < 1 || ac > 2)
	    goto endfunc;
	do_part (source, ac, av);
    }
    else if (!strcmp (cmd, "PRIVMSG") || !strcmp (cmd, "!"))
    {
	char buf[BUFSIZE];

	if (ac != 2)
	    goto endfunc;

	if (validchan(av[0]))
	    goto endfunc;

	if (operserv_on==TRUE)
	    add_msg (source);

	if (!is_ignored (source))
	{
	    if (operserv_on==TRUE)
	    {
		snprintf (buf, sizeof (buf), "%s@%s", s_OperServ, server_name);
		if (stricmp (av[0], s_OperServ) == 0 ||
		    stricmp (av[0], buf) == 0)
		{
		    operserv (source, av[1]);
		    if (!(runflags & RUN_MODE))
			goto endfunc;
		}
	    }
	    if (runflags & RUN_MODE)
	    {
		if (nickserv_on==TRUE)
		{
		    snprintf (buf, sizeof (buf), "%s@%s", s_NickServ, server_name);
		    if (stricmp (av[0], s_NickServ) == 0 ||
			stricmp (av[0], buf) == 0)
			nickserv (source, av[1]);
		}
		if (chanserv_on==TRUE)
		{
		    snprintf (buf, sizeof (buf), "%s@%s", s_ChanServ, server_name);
		    if (stricmp (av[0], s_ChanServ) == 0 ||
			stricmp (av[0], buf) == 0)
			chanserv (source, av[1]);
		}
		if (memoserv_on==TRUE)
		{
		    snprintf (buf, sizeof (buf), "%s@%s", s_MemoServ, server_name);
		    if (stricmp (av[0], s_MemoServ) == 0 ||
			stricmp (av[0], buf) == 0)
			memoserv (source, av[1]);
		}
		if (statserv_on==TRUE)
		{
		    snprintf (buf, sizeof (buf), "%s@%s", s_StatServ, server_name);
		    if (stricmp (av[0], s_StatServ) == 0 ||
			stricmp (av[0], buf) == 0)
			statserv (source, av[1]);
                }
		if (spamchecker_on==TRUE)
		{
		    if (stricmp (av[0], s_SpamChecker) == 0)
			spamchecker (source, av[1]);
		}
		if (infoserv_on==TRUE)
		{
		    snprintf (buf, sizeof (buf), "%s@%s", s_InfoServ, server_name);
		    if (stricmp (av[0], s_InfoServ) == 0 ||
			stricmp (av[0], buf) == 0)
			infoserv (source, av[1]);
		}
	    }
	}
    }
    else if (!strcmp (cmd, "QUIT") || !strcmp (cmd, ","))
    {
	if (ac != 1)
	    goto endfunc;
	do_quit (source, ac, av);
	if (is_services_nick (av[0]))
	    introduce_user (av[0]);
    }
    else if (!strcmp (cmd, "SERVER") || !strcmp (cmd, "'"))
    {
	if (servcnt >= serv_size)
	{
	    if (serv_size < 8)
		serv_size = 8;
	    else
		serv_size *= 2;
	    servlist = srealloc (servlist, sizeof (*servlist) * serv_size);
	}
	servlist[servcnt].server = sstrdup (av[0]);
	servlist[servcnt].uplink = sstrdup (source);
	servlist[servcnt].hops = atoi (av[1]);
	servlist[servcnt].desc = sstrdup (av[2]);
	servlist[servcnt].lag = 999;
	servlist[servcnt].connect = time(NULL);
	servlist[servcnt].version = "Unknown";
	servlist[servcnt].versionflags = "X";
#ifdef IRC_WRECKED
	servlist[servcnt].uptime = time (NULL);
#endif
	servcnt++;
	send_cmd (server_name, "VERSION %s", av[0]);
	send_cmd (server_name, "PING %s %s", server_name, av[0]);
#ifdef IRC_WRECKED
	send_cmd (server_name, "SVSUTIME %s", av[0]);
#endif
	addtimer (sstrdup (av[0]));
    }
    else if (!strcmp (cmd, "STATS") || !strcmp (cmd, "2"))
    {
	if (stricmp (av[0], "u") == 0)
	{
	    int tmpnow;
	    tmpnow = time(NULL) - start_time;

	    send_cmd (server_name, "242 %s :Server Up %d day%s, %d:%02d:%02d",
		   source, tmpnow/86400, (tmpnow/86400 == 1) ? "" : "s",
		   (tmpnow/3600)%24, (tmpnow/60)%60, tmpnow%60);
	    send_cmd (server_name, "250 %s :Highest connection count: 0 (0 clients)", source);
	}
	send_cmd (server_name, "219 %s %s :End of /STATS report", source, av[0]);
    }
    else if (!strcmp (cmd, "TIME") || !strcmp (cmd, ">"))
    {
	time_t t;
	struct tm *tm;
	char buf[80];

	time(&t);
	tm = localtime(&t);
	strftime(buf, sizeof(buf), "%A %B %d %Y -- %H:%M %z", tm);

	send_cmd (server_name, "391 %s %s :%s", source, server_name, buf);
    }
    else if (!strcmp (cmd, "WHOIS") || !strcmp (cmd, "#"))
    {
	if (ac >= 1) {
	    if ((stricmp (av[0], s_NickServ) == 0) && nickserv_on == TRUE)
	    {
		send_cmd (server_name, "311 %s %s %s %s * :%s", source, s_NickServ, ns_user, ns_host, ns_realname);
		send_cmd (server_name, "312 %s %s %s :%s", source, s_NickServ, server_name, server_desc);
		send_cmd (server_name, "317 %s %s 0 %d :seconds idle, signon time", source, s_NickServ, start_time);
	    }
	    else if ((stricmp (av[0], s_ChanServ) == 0) && chanserv_on == TRUE)
	    {
		send_cmd (server_name, "311 %s %s %s %s * :%s", source, s_ChanServ, cs_user, cs_host, cs_realname);
		send_cmd (server_name, "312 %s %s %s :%s", source, s_ChanServ, server_name, server_desc);
		send_cmd (server_name, "317 %s %s 0 %d :seconds idle, signon time", source, s_ChanServ, start_time);
	    }
	    else if ((stricmp (av[0], s_MemoServ) == 0) && memoserv_on == TRUE)
	    {
		send_cmd (server_name, "311 %s %s %s %s * :%s", source, s_MemoServ, ms_user, ms_host, ms_realname);
		send_cmd (server_name, "312 %s %s %s :%s", source, s_MemoServ, server_name, server_desc);
		send_cmd (server_name, "317 %s %s 0 %d :seconds idle, signon time", source, s_MemoServ, start_time);
	    }
	    else if ((stricmp (av[0], s_InfoServ) == 0) && infoserv_on == TRUE)
	    {
		send_cmd (server_name, "311 %s %s %s %s * :%s", source, s_InfoServ, is_user, is_host, is_realname);
		send_cmd (server_name, "312 %s %s %s :%s", source, s_InfoServ, server_name, server_desc);
		send_cmd (server_name, "317 %s %s 0 %d :seconds idle, signon time", source, s_InfoServ, start_time);
	    }
	    else if ((stricmp (av[0], s_GlobalNoticer) == 0) && globalnoticer_on == TRUE)
	    {
		send_cmd (server_name, "311 %s %s %s %s * :%s", source, s_GlobalNoticer, gn_user, gn_host, gn_realname);
		send_cmd (server_name, "312 %s %s %s :%s", source, s_GlobalNoticer, server_name, server_desc);
		send_cmd (server_name, "317 %s %s 0 %d :seconds idle, signon time", source, s_GlobalNoticer, start_time);
	    }
	    else if ((stricmp (av[0], s_StatServ) == 0) && statserv_on == TRUE)
	    {
		send_cmd (server_name, "311 %s %s %s %s * :%s", source, s_StatServ, ss_user, ss_host, ss_realname);
		send_cmd (server_name, "312 %s %s %s :%s", source, s_StatServ, server_name, server_desc);
		send_cmd (server_name, "317 %s %s 0 %d :seconds idle, signon time", source, s_StatServ, start_time);
	    }
	    else if ((stricmp (av[0], s_OperServ) == 0) && operserv_on == TRUE)
	    {
		send_cmd (server_name, "311 %s %s %s %s * :%s", source, s_OperServ, os_user, os_host, os_realname);
		send_cmd (server_name, "312 %s %s %s :%s", source, s_OperServ, server_name, server_desc);
		send_cmd (server_name, "317 %s %s 0 %d :seconds idle, signon time", source, s_OperServ, start_time);
	    }
	    else if ((stricmp (av[0], s_SpamChecker) == 0) && spamchecker_on == TRUE)
	    {
		send_cmd (server_name, "311 %s %s %s %s * :%s", source, s_SpamChecker, sc_user, sc_host, sc_realname);
		send_cmd (server_name, "312 %s %s %s :%s", source, s_SpamChecker, server_name, server_desc);
		send_cmd (server_name, "317 %s %s %li %d :seconds idle, signon time", source, s_SpamChecker,
			1+(int) (70000.0*rand()/(RAND_MAX)+10000.0), start_time);
	    }
	    else
		send_cmd (server_name, "401 %s %s :No such nick/channel", source, av[0]);
	    send_cmd (server_name, "318 %s %s :End of /WHOIS list.", source, av[0]);
	}
    }
    else if (!strcmp (cmd, "SQUIT") || !strcmp (cmd, "-"))
    {
	int i;

	for (i = 0; i < servcnt && stricmp (servlist[i].server, av[0]) != 0; i++);

	if (i < servcnt)
	{
	    if (servlist[i].server)
		free (servlist[i].server);
	    if (servlist[i].desc)
		free (servlist[i].desc);
	    if (servlist[i].uplink)
		free (servlist[i].uplink);
	    if (servlist[i].version)
		free (servlist[i].version);
	    if (servlist[i].versionflags)
		free (servlist[i].versionflags);
	    --servcnt;
	    if (i < servcnt)
		bcopy (servlist + i + 1, servlist + i, sizeof (*servlist) * (servcnt - i));
	}
    }
    else if (!strcmp (cmd, "TOPIC") || !strcmp (cmd, ")"))
    {
	if (ac != 4)
	    goto endfunc;
	do_topic (source, ac, av);
    }
    else if (!strcmp (cmd, "VERSION") || !strcmp (cmd, "+"))
    {
	if (source)
	    send_cmd (server_name, "351 %s paradox-services-%s. %s %s%s%s%s%s%s%s",
		      source, version_number, server_name,
		      (chanserv_on==TRUE)? "C" : "",
                      (nickserv_on==TRUE)? "N" : "",
		      (memoserv_on==TRUE)? "M" : "",
		      (operserv_on==TRUE)? "O" : "",
		      (statserv_on==TRUE)? "S" : "",
		      (spamchecker_on==TRUE)? "s" : "",
                      (runflags & RUN_DEBUG) ? "D" : "");
    }

    endfunc:
    if (av)
	free(av);
}
